// GBA Sprite System - sprite.c
// Nick Sabalausky

#include "gba.h"

#include "userdefs.h"
#include "gfx.h"
#include "sprite.h"

s16 lastTakenCharData = INITIAL_LASTTAKENCHARDATA;

OAMEntry sprites[MAX_SPRITES];
img *spriteImg[MAX_SPRITES];

  // Position
void inline Sprite_SetX(u16 spriteIndex, u16 x)
{
	  // Clamp
	x &= 0x01FF;

	u16 temp = sprites[spriteIndex].attribute1 & 0xFF00;
	sprites[spriteIndex].attribute1 = temp | x;
}

void inline Sprite_SetY(u16 spriteIndex, u16 y)
{
	  // Clamp
	y &= 0x00FF;

	u16 temp = sprites[spriteIndex].attribute0 & 0xFF00;
	sprites[spriteIndex].attribute0 = temp | y;
}

void inline Sprite_SetXY(u16 spriteIndex, u16 x, u16 y)
{
	u16 temp;

	  // Clamp
	x &= 0x01FF;
	y &= 0x00FF;
	
	temp = sprites[spriteIndex].attribute1 & 0xFF00;
	sprites[spriteIndex].attribute1 = temp | x;

	temp = sprites[spriteIndex].attribute0 & 0xFF00;
	sprites[spriteIndex].attribute0 = temp | y;
}

void Sprite_Move(u16 spriteIndex, u16 dx, u16 dy)
{
	  // Get old coords + deltas
	dy += sprites[spriteIndex].attribute0 & 0x00FF;
	dx += sprites[spriteIndex].attribute1 & 0x01FF;

	  // Clamp new coords
	dy &= 0x00FF;
	dx &= 0x01FF;

	  // Clear coordiates
	sprites[spriteIndex].attribute0 &= ~0x00FF;
	sprites[spriteIndex].attribute1 &= ~0x01FF;

	  // Mask in new coords
	sprites[spriteIndex].attribute0 |= dy;
	sprites[spriteIndex].attribute1 |= dx;
}

u16 inline Sprite_GetX(u16 spriteIndex)
{
	return sprites[spriteIndex].attribute1 & 0x01FF;
}

u16 inline Sprite_GetY(u16 spriteIndex)
{
	return sprites[spriteIndex].attribute0 & 0x00FF;
}


  // Horizontal Flip
void inline Sprite_HFlipOn(u16 spriteIndex)
{
	sprites[spriteIndex].attribute1 |= HORIZONTAL_FLIP;
}

void inline Sprite_HFlipOff(u16 spriteIndex)
{
	sprites[spriteIndex].attribute1 &= ~HORIZONTAL_FLIP;
}

void inline Sprite_HFlipToggle(u16 spriteIndex)
{
	sprites[spriteIndex].attribute1 ^= HORIZONTAL_FLIP;
}

  // Vertical Flip
void inline Sprite_VFlipOn(u16 spriteIndex)
{
	sprites[spriteIndex].attribute1 |= VERTICAL_FLIP;
}

void inline Sprite_VFlipOff(u16 spriteIndex)
{
	sprites[spriteIndex].attribute1 &= ~VERTICAL_FLIP;
}

void inline Sprite_VFlipToggle(u16 spriteIndex)
{
	sprites[spriteIndex].attribute1 ^= VERTICAL_FLIP;
}

  // Mosaic
void inline Sprite_MosaicOn(u16 spriteIndex)
{
	sprites[spriteIndex].attribute0 |= MOSAIC;
}

void inline Sprite_MosaicOff(u16 spriteIndex)
{
	sprites[spriteIndex].attribute0 &= ~MOSAIC;
}

void inline Sprite_MosaicToggle(u16 spriteIndex)
{
	sprites[spriteIndex].attribute0 ^= MOSAIC;
}

int  inline Sprite_IsMosaicOn(u16 spriteIndex)
{
	return sprites[spriteIndex].attribute0 & MOSAIC;
}

  // Size Double Flag (If a sprite's rotation flag is turned off,
  //   then turning size double on makes the sprite not get drawn)
void inline Sprite_SizeDoubleOn(u16 spriteIndex)
{
	sprites[spriteIndex].attribute0 |= SIZE_DOUBLE;
}

void inline Sprite_SizeDoubleOff(u16 spriteIndex)
{
	sprites[spriteIndex].attribute0 &= ~SIZE_DOUBLE;
}

void inline Sprite_SizeDoubleToggle(u16 spriteIndex)
{
	sprites[spriteIndex].attribute0 ^= SIZE_DOUBLE;
}

int  inline Sprite_IsSizeDoubleOn(u16 spriteIndex)
{
	return sprites[spriteIndex].attribute0 & SIZE_DOUBLE;
}

  // Character Data
void inline Sprite_SetCharacterOAMOnly(u16 spriteIndex, u16 charIndex)
{
	u16 temp = sprites[spriteIndex].attribute2 & 0xFC00;
	sprites[spriteIndex].attribute2 = temp | charIndex;
}

void inline Sprite_SetCharacter(u16 spriteIndex, img *pImg)
{
	Sprite_SetCharacterOAMOnly(spriteIndex, pImg->charMemIndex);
	spriteImg[spriteIndex] = pImg;
}

void inline Sprite_SetSizeShape(u16 spriteIndex, u16 size, u16 shape)
{
	u16 temp;
	
	temp = sprites[spriteIndex].attribute0 & 0x3FFF;
	sprites[spriteIndex].attribute0 = temp | shape;

	temp = sprites[spriteIndex].attribute1 & 0x3FFF;
	sprites[spriteIndex].attribute1 = temp | size;
}

void inline Sprite_SetColorDepth(u16 spriteIndex, enumSpriteColorDepth colorDepth)
{
	u16 temp = sprites[spriteIndex].attribute0 & ~(1<<13);
	sprites[spriteIndex].attribute0 = temp | colorDepth;

	if(colorDepth == SPRITE_4BIT)
	{
		  // Set palette to #0
		sprites[spriteIndex].attribute2 = sprites[spriteIndex].attribute2 & 0x0FFF;
	}
}

void inline LoadSpriteImage8BitTo(u16 charIndex, img *pImg)
{
	u16 numPixels = pImg->width * pImg->height;
	const u16 *data = pImg->data;

	int index;
	for(index=0; index < numPixels/2; index++)
	{
		CharMem[(charIndex*8*2) + index] = data[index];
	}

	pImg->charMemIndex = charIndex;
}

void LoadSpriteImage(img *pImg, enumSpriteColorDepth colorDepth)
{
	u16 numPixels = pImg->width * pImg->height;
	int numTiles = (numPixels / (8*8));
	s16 charIndex;

	if(colorDepth == SPRITE_8BIT)
		numTiles *= 2;

	charIndex = lastTakenCharData - numTiles;

	u16 temp = REG_DISPCNT_L; // REG_DISPCNT_L is volitile, need to work w/ a copy
	if((temp & 0x0007) >= 3) // If in screen mode 3, 4, or 5
	{
		if(lastTakenCharData < 512)
		{
			pImg->charMemIndex = -1;
			return;
		}
	}
	else
	{
		if(lastTakenCharData < 0)
		{
			pImg->charMemIndex = -1;
			return;
		}
	}

	int index;
	const u16 *data = pImg->data;
	int indexMax = numPixels/4;
	if(colorDepth == SPRITE_8BIT)
		indexMax *= 2;

	for(index=0; index < indexMax; index++)
	{
		CharMem[(charIndex*8*2) + index] = data[index];
	}

	lastTakenCharData = charIndex;
	pImg->charMemIndex = charIndex;
}

void inline ResetSpriteImageData()
{
	lastTakenCharData = INITIAL_LASTTAKENCHARDATA;
}

void inline LoadSpritePalette8Bit(const u16 *data)
{
	int index;
	for(index=0; index<256; index++)
	{
		OBJPaletteMem[index] = data[index];
	}
}

void inline SetSpritePaletteEntry8Bit(u16 index, u16 color)
{
	OBJPaletteMem[index] = color;
}

u16 inline GetSpritePaletteEntry8Bit(u16 index)
{
	return OBJPaletteMem[index];
}

  // Update Sprite -> OAMSprite
void Sprite_UpdateOAM(int oamIndexStart, int numEntries)
{
	  // Slow:
//	int index;
//	for(index=0; index < numEntries; index++)
//	{
//		OAMmem[(oamIndexStart+index)*4 + 0] = sprites[index].attribute0;
//		OAMmem[(oamIndexStart+index)*4 + 1] = sprites[index].attribute1;
//		OAMmem[(oamIndexStart+index)*4 + 2] = sprites[index].attribute2;
//	}

	  // Faster
	OAMEntry* oamDestPtr = (OAMEntry*)OAMmem + oamIndexStart;
	OAMEntry* oamDestEndCopy = oamDestPtr + numEntries;

	OAMEntry* oamSrcPtr = sprites;
	
	for(; oamDestPtr < oamDestEndCopy; oamDestPtr++, oamSrcPtr++)
	{
		*oamDestPtr = *oamSrcPtr;
	}
}

void Sprite_ClearOAM(int oamIndexStart, int numEntries)
{
	  // Slow:
//	int index;
//	for(index=0; index < numEntries; index++)
//	{
//		OAMmem[(oamIndexStart+index)*4 + 0] = sprites[index].attribute0;
//		OAMmem[(oamIndexStart+index)*4 + 1] = sprites[index].attribute1;
//		OAMmem[(oamIndexStart+index)*4 + 2] = sprites[index].attribute2;
//	}

	  // Faster
	OAMEntry* oamDestPtr = (OAMEntry*)OAMmem + oamIndexStart;
	OAMEntry* oamShadowPtr = sprites + oamIndexStart;
	OAMEntry* oamDestEndCopy = oamDestPtr + numEntries;

	OAMEntry oamSrc = {0x0000, 0x0000, 0x0000, 0x0000};
	
	for(; oamDestPtr < oamDestEndCopy; oamDestPtr++, oamShadowPtr++)
	{
		*oamDestPtr = oamSrc;
		*oamShadowPtr = oamSrc;
	}
}

void Sprite_InitXY(u16 spriteIndex, enumSpriteColorDepth colorDepth, u16 x, u16 y)
{
	img *pImg = spriteImg[spriteIndex];

	Sprite_SetCharacterOAMOnly(spriteIndex, pImg->charMemIndex);
	Sprite_SetColorDepth(spriteIndex, colorDepth);
	Sprite_SetSizeShape(spriteIndex, pImg->size, pImg->shape);
	Sprite_SetXY(spriteIndex, x, y);
}

// END OF FILE //////////////////////////
